from decimal import Decimal


def fnumber(value: Decimal, separator: str, extra: str | None = None) -> str:
    str_value = f"{value:,}".replace(",", separator)
    if extra == "$":
        return "$" + str_value
    elif extra == "%":
        return str_value + "%"
    return str_value



from mm_std import print_json

from mm_balance.config import Config
from mm_balance.price import Prices
from mm_balance.result import BalancesResult
from mm_balance.token_decimals import TokenDecimals
from mm_balance.workers import Workers


def print_result(config: Config, token_decimals: TokenDecimals, prices: Prices, workers: Workers, result: BalancesResult) -> None:
    data: dict[str, object] = {}
    if config.print_debug:
        data["nodes"] = config.nodes
        data["token_decimals"] = token_decimals
    if config.price:
        data["prices"] = prices

    data["groups"] = result.groups
    data["total"] = result.total
    if config.has_share():
        data["total_share"] = result.total_share

    errors = workers.get_errors()
    if errors:
        data["errors"] = errors

    print_json(data)



from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

RETRIES_BALANCE = 5
RETRIES_DECIMALS = 5
RETRIES_COINGECKO_PRICES = 5
TIMEOUT_BALANCE = 5
TIMEOUT_DECIMALS = 5


class Network(str):
    def is_evm_network(self) -> bool:
        return self in [NETWORK_ETHEREUM, NETWORK_ARBITRUM_ONE, NETWORK_OP_MAINNET] or self.startswith("evm-")

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))


NETWORK_APTOS = Network("aptos")
NETWORK_ARBITRUM_ONE = Network("arbitrum-one")
NETWORK_BITCOIN = Network("bitcoin")
NETWORK_ETHEREUM = Network("ethereum")
NETWORK_SOLANA = Network("solana")
NETWORK_OP_MAINNET = Network("op-mainnet")
NETWORKS = [NETWORK_APTOS, NETWORK_ARBITRUM_ONE, NETWORK_BITCOIN, NETWORK_ETHEREUM, NETWORK_SOLANA, NETWORK_OP_MAINNET]


TOKEN_ADDRESS: dict[Network, dict[str, str]] = {
    NETWORK_ETHEREUM: {
        "USDT": "0xdac17f958d2ee523a2206206994597c13d831ec7",
        "USDC": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
    },
    NETWORK_SOLANA: {
        "USDT": "Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB",
        "USDC": "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v",
    },
    NETWORK_ARBITRUM_ONE: {
        "USDT": "0xfd086bc7cd5c481dcc9c85ebe478a1c0b69fcbb9",
        "USDC": "0xff970a61a04b1ca14834a43f5de4533ebddb5cc8",
    },
    NETWORK_OP_MAINNET: {
        "USDT": "0x94b008aA00579c1307B0EF2c499aD98a8ce58e58",
        "USDC": "0x7f5c764cbc14f9669b88837ca1490cca17c31607",
    },
}

TICKER_TO_COINGECKO_ID = {
    "BTC": "bitcoin",
    "ETH": "ethereum",
    "USDT": "tether",
    "USDC": "usd-coin",
    "SOL": "solana",
    "APT": "aptos",
    "POL": "matic-network",
}

USD_STABLECOINS = ["USDT", "USDC"]

DEFAULT_NODES: dict[Network, list[str]] = {
    NETWORK_ARBITRUM_ONE: ["https://arb1.arbitrum.io/rpc", "https://arbitrum.llamarpc.com"],
    NETWORK_BITCOIN: [],
    NETWORK_ETHEREUM: ["https://ethereum.publicnode.com", "https://rpc.ankr.com/eth"],
    NETWORK_SOLANA: ["https://api.mainnet-beta.solana.com"],
    NETWORK_OP_MAINNET: ["https://mainnet.optimism.io", "https://optimism.llamarpc.com"],
    NETWORK_APTOS: ["https://fullnode.mainnet.aptoslabs.com/v1"],
}



from decimal import Decimal

from mm_aptos import balance
from mm_std import Result

from mm_balance.constants import RETRIES_BALANCE, TIMEOUT_BALANCE


def get_balance(
    nodes: list[str], wallet: str, token: str | None, decimals: int, proxies: list[str], round_ndigits: int
) -> Result[Decimal]:
    if token is None:
        token = "0x1::aptos_coin::AptosCoin"  # nosec
    return balance.get_decimal_balance_with_retries(
        RETRIES_BALANCE,
        nodes,
        wallet,
        token,
        decimals=decimals,
        timeout=TIMEOUT_BALANCE,
        proxies=proxies,
        round_ndigits=round_ndigits,
    )



from collections import defaultdict
from dataclasses import dataclass
from decimal import Decimal

from mm_std import Ok

from mm_balance.config import Config, Group
from mm_balance.constants import USD_STABLECOINS, Network
from mm_balance.price import Prices
from mm_balance.workers import Task, Workers


@dataclass
class Balance:
    balance: Decimal
    usd_value: Decimal  # 0 if config.price is False


@dataclass
class AddressBalance:
    address: str
    balance: Balance | str  # balance value or error message


@dataclass
class GroupResult:
    ticker: str
    network: Network
    comment: str
    share: Decimal
    addresses: list[AddressBalance]
    balance_sum: Decimal  # sum of all balances in the group
    usd_sum: Decimal  # sum of all usd values in the group
    balance_sum_share: Decimal  # sum of all balances in the group multiplied by share
    usd_sum_share: Decimal  # sum of all usd values in the group multiplied by share


@dataclass
class Total:
    coin_balances: dict[str, Decimal]
    coin_usd_values: dict[str, Decimal]
    portfolio_share: dict[str, Decimal]  # ticker -> usd value % from total usd value
    stablecoin_sum: Decimal  # sum of usd stablecoins: usdt, usdc, etc..
    total_usd_sum: Decimal  # sum of all coins in USD


@dataclass
class BalancesResult:
    groups: list[GroupResult]
    total: Total
    total_share: Total


def create_balances_result(config: Config, prices: Prices, workers: Workers) -> BalancesResult:
    groups = []
    for group_index, group in enumerate(config.groups):
        tasks = workers.get_group_tasks(group_index, group.network)
        groups.append(_create_group_result(config, group, tasks, prices))

    total = _create_total(False, groups)
    total_share = _create_total(True, groups)
    return BalancesResult(groups=groups, total=total, total_share=total_share)


def _create_total(use_share: bool, groups: list[GroupResult]) -> Total:
    coin_balances: dict[str, Decimal] = defaultdict(Decimal)  # ticker -> balance
    coin_usd_values: dict[str, Decimal] = defaultdict(Decimal)  # ticker -> usd value
    portfolio_share: dict[str, Decimal] = defaultdict(Decimal)  # ticker -> usd value % from total usd value
    total_usd_sum = Decimal(0)
    stablecoin_sum = Decimal(0)

    for group in groups:
        balance_value = group.balance_sum_share if use_share else group.balance_sum
        usd_value = group.usd_sum_share if use_share else group.usd_sum
        coin_balances[group.ticker] += balance_value
        coin_usd_values[group.ticker] += usd_value
        if group.ticker in USD_STABLECOINS:
            stablecoin_sum += usd_value  # TODO: or balance_value?
        total_usd_sum += usd_value

    if total_usd_sum > 0:
        for ticker, usd_value in coin_usd_values.items():
            if ticker in USD_STABLECOINS:
                portfolio_share[ticker] = round(stablecoin_sum * 100 / total_usd_sum, 2)
            else:
                portfolio_share[ticker] = round(usd_value * 100 / total_usd_sum, 2)

    return Total(
        coin_balances=coin_balances,
        coin_usd_values=coin_usd_values,
        portfolio_share=portfolio_share,
        stablecoin_sum=stablecoin_sum,
        total_usd_sum=total_usd_sum,
    )


def _create_group_result(config: Config, group: Group, tasks: list[Task], prices: Prices) -> GroupResult:
    addresses = []
    balance_sum = Decimal(0)
    usd_sum = Decimal(0)
    for task in tasks:
        balance: Balance | str
        if task.balance is None:
            balance = "balance is None! Something went wrong."
        else:
            if isinstance(task.balance, Ok):
                coin_value = task.balance.ok
                usd_value = Decimal(0)
                if group.ticker in prices:
                    usd_value = round(coin_value * prices[group.ticker], config.round_ndigits)
                balance = Balance(balance=coin_value, usd_value=usd_value)
                balance_sum += balance.balance
                usd_sum += balance.usd_value
            else:
                balance = task.balance.err
        addresses.append(AddressBalance(address=task.wallet_address, balance=balance))

    balance_sum_share = balance_sum * group.share
    usd_sum_share = usd_sum * group.share

    return GroupResult(
        ticker=group.ticker,
        network=group.network,
        comment=group.comment,
        share=group.share,
        addresses=addresses,
        balance_sum=balance_sum,
        usd_sum=usd_sum,
        balance_sum_share=balance_sum_share,
        usd_sum_share=usd_sum_share,
    )



import getpass
import pathlib
import pkgutil
from typing import Annotated

import typer
from mm_std import PrintFormat, fatal

from mm_balance.config import Config
from mm_balance.constants import NETWORKS
from mm_balance.output.formats import json_format, table_format
from mm_balance.price import Prices, get_prices
from mm_balance.result import create_balances_result
from mm_balance.token_decimals import get_token_decimals
from mm_balance.workers import Workers

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False, add_completion=False)


def example_callback(value: bool) -> None:
    if value:
        data = pkgutil.get_data(__name__, "config/example.yml")
        typer.echo(data)
        raise typer.Exit


def networks_callback(value: bool) -> None:
    if value:
        for network in NETWORKS:
            typer.echo(network)
        raise typer.Exit


@app.command()
def cli(
    config_path: Annotated[pathlib.Path, typer.Argument()],
    print_format: Annotated[PrintFormat | None, typer.Option("--format", "-f", help="Print format.")] = None,
    skip_empty: Annotated[bool | None, typer.Option("--skip-empty", "-s", help="Skip empty balances.")] = None,
    debug: Annotated[bool | None, typer.Option("--debug", "-d", help="Print debug info.")] = None,
    price: Annotated[bool | None, typer.Option("--price/--no-price", help="Print prices.")] = None,
    _example: Annotated[bool | None, typer.Option("--example", callback=example_callback, help="Print a config example.")] = None,
    _networks: Annotated[
        bool | None, typer.Option("--networks", callback=networks_callback, help="Print supported networks.")
    ] = None,
) -> None:
    zip_password = ""  # nosec
    if config_path.name.endswith(".zip"):
        zip_password = getpass.getpass("zip password")
    config = Config.read_config(config_path, zip_password=zip_password)

    if print_format is not None:
        config.print_format = print_format
    if debug is not None:
        config.print_debug = debug
    if skip_empty is not None:
        config.skip_empty = skip_empty
    if price is not None:
        config.price = price

    if config.print_debug and config.print_format is PrintFormat.TABLE:
        table_format.print_nodes(config)

    token_decimals = get_token_decimals(config)
    if config.print_debug and config.print_format is PrintFormat.TABLE:
        table_format.print_token_decimals(token_decimals)

    prices = get_prices(config) if config.price else Prices()
    if config.print_format is PrintFormat.TABLE:
        table_format.print_prices(config, prices)

    workers = Workers(config, token_decimals)
    workers.process()

    result = create_balances_result(config, prices, workers)
    if config.print_format is PrintFormat.TABLE:
        table_format.print_result(config, result, workers)
    elif config.print_format is PrintFormat.JSON:
        json_format.print_result(config, token_decimals, prices, workers, result)
    else:
        fatal("Unsupported print format")


if __name__ == "__main__":
    app()



from mm_std import Err, fatal

from mm_balance.config import Config
from mm_balance.constants import NETWORK_APTOS, NETWORK_BITCOIN, NETWORK_SOLANA, Network
from mm_balance.rpc import evm, solana


class TokenDecimals(dict[Network, dict[str | None, int]]):  # {network: {None: 18}} -- None is for native token, ex. ETH
    def __init__(self, networks: list[Network]) -> None:
        super().__init__()
        for network in networks:
            self[network] = {}


def get_token_decimals(config: Config) -> TokenDecimals:
    result = TokenDecimals(config.networks())
    proxies = config.proxies

    for group in config.groups:
        # token_decimals is already known
        if group.token_decimals is not None:
            result[group.network][group.token_address] = group.token_decimals
            continue

        # get token_decimals for known native tokens
        if group.token_address is None:
            if group.network.is_evm_network():
                result[group.network][None] = 18
            elif group.network == NETWORK_SOLANA:
                result[group.network][None] = 9
            elif group.network == NETWORK_BITCOIN:
                result[group.network][None] = 8
            elif group.network == NETWORK_APTOS:
                result[group.network][None] = 8
            else:
                fatal(f"Can't get token decimals for native token on network: {group.network}")
            continue

        # get token_decimals via RPC
        # TODO: group.token_address must be in normalized form, otherwise it can be different for the same token
        if group.token_address in result[group.network]:
            continue  # don't request for a token_decimals twice

        nodes = config.nodes[group.network]
        if group.network.is_evm_network():
            res = evm.get_token_decimals(nodes, group.token_address, proxies)
        elif group.network == NETWORK_SOLANA:
            res = solana.get_token_decimals(nodes, group.token_address, proxies)
        else:
            fatal(f"unsupported network: {group.network}. Cant get token decimals for {group.token_address}")
        if isinstance(res, Err):
            fatal(f"can't get decimals for token {group.ticker} / {group.token_address}, error={res.err}")
        result[group.network][group.token_address] = res.ok

    return result



from decimal import Decimal

from rich.progress import BarColumn, MofNCompleteColumn, Progress, TaskID, TextColumn


def format_number(value: Decimal, separator: str, extra: str | None = None) -> str:
    str_value = f"{value:,}".replace(",", separator)
    if extra == "$":
        return "$" + str_value
    elif extra == "%":
        return str_value + "%"
    return str_value


def create_progress_bar(disable: bool) -> Progress:
    return Progress(TextColumn("[progress.description]{task.description}"), BarColumn(), MofNCompleteColumn(), disable=disable)


def create_progress_task(progress: Progress, description: str, total: int) -> TaskID:
    return progress.add_task("[green]" + description, total=total)



from decimal import Decimal

from mm_eth import erc20, rpc
from mm_std import Ok, Result

from mm_balance.constants import RETRIES_BALANCE, RETRIES_DECIMALS, TIMEOUT_BALANCE, TIMEOUT_DECIMALS


def get_balance(
    nodes: list[str], wallet: str, token: str | None, decimals: int, proxies: list[str], round_ndigits: int
) -> Result[Decimal]:
    if token is not None:
        res = erc20.get_balance(
            nodes,
            token,
            wallet,
            proxies=proxies,
            attempts=RETRIES_BALANCE,
            timeout=TIMEOUT_BALANCE,
        )
    else:
        res = rpc.eth_get_balance(nodes, wallet, proxies=proxies, attempts=RETRIES_BALANCE, timeout=TIMEOUT_BALANCE)
    return res.and_then(lambda b: Ok(round(Decimal(b / 10**decimals), round_ndigits)))


def get_token_decimals(nodes: list[str], token_address: str, proxies: list[str]) -> Result[int]:
    return erc20.get_decimals(nodes, token_address, timeout=TIMEOUT_DECIMALS, proxies=proxies, attempts=RETRIES_DECIMALS)



from dataclasses import dataclass
from decimal import Decimal

from mm_std import ConcurrentTasks, PrintFormat, Result
from rich.progress import TaskID

from mm_balance.config import Config
from mm_balance.constants import NETWORK_APTOS, NETWORK_BITCOIN, NETWORK_SOLANA, Network
from mm_balance.output import utils
from mm_balance.rpc import aptos, btc, evm, solana
from mm_balance.token_decimals import TokenDecimals


@dataclass
class Task:
    group_index: int
    wallet_address: str
    token_address: str | None
    balance: Result[Decimal] | None = None


class Workers:
    def __init__(self, config: Config, token_decimals: TokenDecimals):
        self.config = config
        self.token_decimals = token_decimals
        self.tasks: dict[Network, list[Task]] = {network: [] for network in config.networks()}
        self.progress_bar = utils.create_progress_bar(config.print_format is not PrintFormat.TABLE)
        self.progress_bar_task: dict[Network, TaskID] = {}

        for idx, group in enumerate(config.groups):
            task_list = [Task(group_index=idx, wallet_address=a, token_address=group.token_address) for a in group.addresses]
            self.tasks[group.network].extend(task_list)

        for network in config.networks():
            if self.tasks[network]:
                self.progress_bar_task[network] = utils.create_progress_task(self.progress_bar, network, len(self.tasks[network]))

    def process(self) -> None:
        with self.progress_bar:
            job = ConcurrentTasks(max_workers=10)
            for network in self.config.networks():
                job.add_task(network, self._process_network, args=(network,))
            job.execute()

    def get_group_tasks(self, group_index: int, network: Network) -> list[Task]:
        # TODO: can we get network by group_index?
        return [b for b in self.tasks[network] if b.group_index == group_index]

    def get_errors(self) -> list[Task]:
        result = []
        for network in self.tasks:
            for task in self.tasks[network]:
                if task.balance is not None and task.balance.is_err():
                    result.append(task)
        return result

    def _process_network(self, network: Network) -> None:
        job = ConcurrentTasks(max_workers=self.config.workers[network])
        for idx, task in enumerate(self.tasks[network]):
            job.add_task(str(idx), self._get_balance, args=(network, task.wallet_address, task.token_address))
        job.execute()
        # TODO: print job.exceptions if present
        for idx, _task in enumerate(self.tasks[network]):
            self.tasks[network][idx].balance = job.result.get(str(idx))  # type: ignore[assignment]

    def _get_balance(self, network: Network, wallet_address: str, token_address: str | None) -> Result[Decimal]:
        nodes = self.config.nodes[network]
        round_ndigits = self.config.round_ndigits
        proxies = self.config.proxies
        token_decimals = self.token_decimals[network][token_address]

        if network.is_evm_network():
            res = evm.get_balance(nodes, wallet_address, token_address, token_decimals, proxies, round_ndigits)
        elif network == NETWORK_BITCOIN:
            res = btc.get_balance(wallet_address, proxies, round_ndigits)
        elif network == NETWORK_APTOS:
            res = aptos.get_balance(nodes, wallet_address, token_address, token_decimals, proxies, round_ndigits)
        elif network == NETWORK_SOLANA:
            res = solana.get_balance(nodes, wallet_address, token_address, token_decimals, proxies, round_ndigits)
        else:
            raise ValueError(f"Unsupported network: {network}")

        self.progress_bar.update(self.progress_bar_task[network], advance=1)
        return res
